////////////////////////////////////////////////////////////////////////////////
//	
//	COPYRIGHT 2003 BY RADISYS CORPORATION.  ALL RIGHTS RESERVED
//
////////////////////////////////////////////////////////////////////////////////
//	pkt_echo.uc
//
//	This file sources packets from the packet_rx[] microblock and forwards them
//	to the corresponding port transmit queue for the sphy_mphy4_tx[] microblock.
//
////////////////////////////////////////////////////////////////////////////////

#include <dl_system.h>
#include <stdmac.uc>
#include <dispatch_loop.uc>
#include <hardware.h>



#define NUM_WORDS_TX_MESSAGE	1



////////////////////////////////////////////////////////////////////////////////
//
//	Signal and register definitions
//
////////////////////////////////////////////////////////////////////////////////

.sig 			sig_scr_get			; signal for scratch get
.sig volatile	sig_next			; signal for next ctx thread ordering
.sig			sig_scr_put			; signal for scratch put

.reg 	temp				; general purpose use
.reg	ring				; 
.reg	port				; input port number
.reg	$txreq				; tx request message to put on scratch rings
.reg	zero				; gpr containing value 0
.reg	one					; gpr containing value 1
.reg	sig_next_val		; value to write to same_me_signal register
.reg 	sig_mask
.reg	ctx_num				; context number of the thread
.reg	port_out			; output port to forward packet to

.reg	@count_p0
.reg	@count_p1
.reg	@count_p2
.reg	@count_p3

.reg	@count_empty

.reg	@debug_cnt

xbuf_alloc[$rdata, RX_TO_FUNC_MSG_SIZE, read]			; allocate xfer registers


// This is a global signal that indicates that 
// system initialization is done

.sig	volatile system_init_done
.addr 	system_init_done				me_init_signal



////////////////////////////////////////////////////////////////////////////////
//
//	Context common initializations
//
////////////////////////////////////////////////////////////////////////////////

immed[zero, 0]
immed[one, 1]
alu[sig_next_val, --, B, &sig_next, <<3]

// get context num
local_csr_rd[active_ctx_sts]
immed[ctx_num,0] 	//immed operand contains local_csr_rd data
alu[ctx_num, ctx_num, AND, 0x07]



////////////////////////////////////////////////////////////////////////////////
//
//	Context specific initializations
//
////////////////////////////////////////////////////////////////////////////////

.if( ctx() == 0 )
	alu[sig_next_val, sig_next_val, OR, 1]
	alu[@count_p0, --, b, zero]
	alu[@count_p1, --, b, zero]
	alu[@count_p2, --, b, zero]
	alu[@count_p3, --, b, zero]
	alu[@count_empty, --, b, zero]
	alu[@debug_cnt, --, b, zero]

.elif( ctx() == 1 )
	alu[sig_next_val, sig_next_val, OR, 2]

.elif( ctx() == 2 )
	alu[sig_next_val, sig_next_val, OR, 3]

.elif( ctx() == 3 )
	alu[sig_next_val, sig_next_val, OR, 4]

.elif( ctx() == 4 )
	alu[sig_next_val, sig_next_val, OR, 5]

.elif( ctx() == 5 )
	alu[sig_next_val, sig_next_val, OR, 6]

.elif( ctx() == 6 )
	alu[sig_next_val, sig_next_val, OR, 7]

.elif( ctx() == 7 )
	alu[sig_next_val, sig_next_val, OR, 0]

.endif


//	All threads wait for system go signal

.if( ctx() == 0 )

	// this is the signal that all blocks are to 
	// wait on to indicate that system initialization 
	// is done
		
	ctx_arb[system_init_done]

	// signal all contexts to proceed

	alu[temp, --, B, &sig_next, <<3]
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 1
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 2
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 3
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 4
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 5
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 6
	alu[temp, temp, +, 1]
	local_csr_wr[SAME_ME_SIGNAL, temp]	//thread 7
	alu[temp, temp, +, 1]

	nop
	nop

.else

	.set_sig sig_next
	ctx_arb[sig_next]

.endif


// kick start next thread sequencing by thread 0 signaling itself

.if( ctx() == 0 )
	
	alu[temp, --, b, &sig_next, <<3]
	local_csr_wr[SAME_ME_SIGNAL, temp]

	nop
	nop
	nop

.endif


// Kick start scr_put signals 

alu[temp, ctx_num, OR, &sig_scr_put, <<3]
local_csr_wr[SAME_ME_SIGNAL, temp]
nop
nop
nop

// Tell compiler that sig_scr_put has been raised for first time through loop

.set_sig sig_scr_put




////////////////////////////////////////////////////////////////////////////////
//
//	Main loop
//
////////////////////////////////////////////////////////////////////////////////

start#:

	// Tell compiler that sig_next is raised each time through loop
	.set_sig sig_next

	// Wait for signal from previous thread and scr_put from prior loop to 
	// ensure thread ordering
w1#:	;label for debug
	ctx_arb[sig_next, sig_scr_put]

	// Source packet from pkt rx scratch ring
	alu_shf[ring, --, B, POS_RX_RING_OUT, <<2]
	scratch[get, $rdata0, 0, ring, RX_TO_FUNC_MSG_SIZE], sig_done[sig_scr_get]

	// Signal next thread to go
s1#:	;label for debug
	local_csr_wr[SAME_ME_SIGNAL, sig_next_val]
;	nop
;	nop
;	nop

	// Wait on scratch read 
w2#:	;label for debug
	ctx_arb[sig_scr_get]


	// check for empty ring
	alu[--, $rdata0, -, 0]
	beq[ring_empty#]

	// if ring not empty

		// Extract input port
		alu[port, 0xF, AND, $rdata4, >>16]

		// Calculate output port (0->1, 1->2, 2->0)
		alu[port_out, port, +, 1]
		.if(port_out == 3)
			alu[port_out, --, b, 0]
		.endif

		// Build tx request for transmit queue	
		alu[temp, --, b, port_out, <<24]					; 27:24 = output port
		ld_field[temp, 0111, $rdata0]						; 23:00 = ptr to sop buffer desc
		alu[$txreq, temp, OR, 1, <<31] 						; 31    = set valid bit (31:28 = reserved )
	
		// Jump to scratch ring write for each port
		alu[temp, --, b, port_out, <<2]
		jump[temp, write_ring0#], targets[write_ring0#,write_ring1#,write_ring2#,write_ring3#]
	
		write_ring0#:
			br_inp_state[SCR_RING/**/PACKET_TX_SCR_RING_0/**/_FULL, full_ring0#]
			scratch[put, $txreq, zero, (PACKET_TX_SCR_RING_0<<2), NUM_WORDS_TX_MESSAGE], sig_done[sig_scr_put]
			alu[@count_p0, @count_p0, +, one]
			br[start#]
		write_ring1#:
			br_inp_state[SCR_RING/**/PACKET_TX_SCR_RING_1/**/_FULL, full_ring1#]
			scratch[put, $txreq, zero, (PACKET_TX_SCR_RING_1<<2), NUM_WORDS_TX_MESSAGE], sig_done[sig_scr_put]
			alu[@count_p1, @count_p1, +, one]
			br[start#]
		write_ring2#:
			br_inp_state[SCR_RING/**/PACKET_TX_SCR_RING_2/**/_FULL, full_ring2#]
			scratch[put, $txreq, zero, (PACKET_TX_SCR_RING_2<<2), NUM_WORDS_TX_MESSAGE], sig_done[sig_scr_put]
			alu[@count_p2, @count_p2, +, one]
			br[start#]
		write_ring3#:
			br_inp_state[SCR_RING/**/PACKET_TX_SCR_RING_3/**/_FULL, full_ring3#]
			scratch[put, $txreq, zero, (PACKET_TX_SCR_RING_3<<2), NUM_WORDS_TX_MESSAGE], sig_done[sig_scr_put]
			alu[@count_p3, @count_p3, +, one]
			br[start#]

		
		full_ring0#:
			ctx_arb[voluntary]
			br[write_ring0#]

		full_ring1#:
			ctx_arb[voluntary]
			br[write_ring1#]

		full_ring2#:
			ctx_arb[voluntary]
			br[write_ring2#]
		
		full_ring3#:
			ctx_arb[voluntary]
			br[write_ring3#]



ring_empty#:

		// raise sig_scr_put signal
		alu[temp, ctx_num, OR, &sig_scr_put, <<3]
		local_csr_wr[SAME_ME_SIGNAL, temp]
		alu[@count_empty, @count_empty, +, one]
;		nop
;		nop
;		nop
		.set_sig sig_scr_put	//tell assembler signal has been raised		

br[start#]
