///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

#macro dl_source_pkt_rx[ io_sig_mask ]
.begin

	.sig scr_get											; signal for scratch get
	.reg tmp

	xbuf_alloc[$rdata, RX_TO_FUNC_MSG_SIZE, read]			; allocate xfer registers

	// Get Data from Scratch Ring.
	alu_shf[ring, --, B, POS_RX_RING_OUT, <<2]
	scratch[get, $rdata0, 0, ring, RX_TO_FUNC_MSG_SIZE], sig_done[scr_get]

	// Add the signal to the signal mask on which to wait
	alu[tmp, --, B, &scr_get]
	alu[--, tmp, OR, 0]										; tmp will be used  by "indirect" in next instruction
	alu_shf[io_sig_mask, io_sig_mask, OR, 1, <<indirect]	; shift 1 by scr_get times

	// Check if Ring is empty. ==> it would have returned zero
	alu[--, $rdata0, -, 0]									; Check for 0. i.e check if ring empty
	beq[ring_empty#]								

	//	$rdata0...$rdata4 will have valid data now.
	//	Load Meta Data Cache with $rdata2, and $rdata3 << 16
	// 	$rdata2 (buffer size,offset) => LW1 of meta data
	//	$rdata3(pkt size, rx stat, free_list etc) ==> LW2 in meta data. 
	//	LSB16 bits of $rdata4 is inport => LW3, MSB16 of meta data

	dl_meta_init_cache[0, $rdata2, $rdata3, $rdata4, 0, 0, 0, 0]
    dl_meta_set_nexthop_id[0xffff]	// set nexthop ID to -1
    alu[dl_exception_reg, --, b, 0]	/* reset, otherwise may skewup core
  									   component by residual value */

	// 	We branch here so that we can take advantage of the defer slots here.
	//	This branch with defer, will finally save a branch to the end of this macro 
	//	(where 2 cycles may be wasted)

	br[load_ip_cache#], defer[2]
	move[dl_buf_handle, $rdata0]				; LW0 from ring is the current buffer handle (containign SOP)
	move[dl_eop_buf_handle, $rdata1]			; LW1 from ring is last buffer handle (for large packets)

	// 	program flow continues in load_ip_cache#:

	//	Handle Ring empty Condition.
ring_empty#:

	//	If ring is empty, just set dl_buf_handle and dl_next_block to IX_NULL;
	// 	Is this the right thing to do ? TBD ???

	move[dl_buf_handle, IX_NULL]				; Note IX_NULL is NOT ZERO.
	move[dl_next_block, IX_NULL]
	br[nothing_to_do#]

	move[dl_next_block, DL_SOURCE_NEXT1]

nothing_to_do#:

	xbuf_free[$rdata]							; Free xfer registers. No Code generated.



.end
#endm



