
#define  IXP_MICROC

#include "dl_system.h"
#include <hardware.h>
#include <intrinsic.c>
#include <libc.c>
#include <rtl.c>
#include <stdlib.h>
#include <ixptypes.h>
#include <string.h>
#include <ixp.h>
#include <ixp_lib.h>
#include <dl_meta.h>
#include <dl_buf.c>

#define NUM_WORDS_TX_MESSAGE	1
#define 	zero 	0
#define 	one		1

typedef struct
{
	dl_buf_handle_t		dlBufHandle;			// lw0
	unsigned int 		buffer_next		: 32;	// lw1
	unsigned int		buffer_size		: 16;	// lw2
	unsigned int 		offset			: 16;
	unsigned int		packet_size		: 16;	// lw3
	unsigned int		free_list_id	: 4;
	unsigned int		rx_stat			: 4;
	unsigned int 		header_type		: 8;
	unsigned int		input_port		: 16;	// lw4
	unsigned int		output_port		: 16;
} ring_data_t3;

typedef struct 
{
	//QW0
	unsigned int dest_MAC_high32;
	unsigned int dest_MAC_low16   :16;
	unsigned int source_MAC_high16:16;
	//QW1
	unsigned int source_MAC_low32;
	unsigned int pdu			  :16;
	unsigned int version   		  :4;
	unsigned int header_len       :4;
	unsigned int tos	   		  :8;
	//QW2
	unsigned int total_len		  :16;
	unsigned int id				  :16;
	unsigned int flag			  :4;
	unsigned int frag_offset	  :12;
	unsigned int ttl			  :8;
	unsigned int protocol		  :8;
	//QW3
	unsigned int checksum		  :16;
//	unsigned int source_addr;
	unsigned int source_addr_high	:16;
	unsigned int source_addr_low	:16;
	unsigned int destination_addr_high:16;
	//QW4
	unsigned int destination_addr_low:16;
	unsigned int source_port      :16;  // UDP and TCP only
	unsigned int destination_port :16;  // UDP and TCP only
	unsigned int reserver_1		  :16;
	//QW5
	unsigned long long	reserver_2;
} ip_header;

typedef struct
{
	unsigned int sa;
	unsigned int da;
	unsigned short sp_begin;
	unsigned short sp_end;
	unsigned short dp_begin;
	unsigned short dp_end;
	unsigned char sa_len;
	unsigned char da_len;
	unsigned char protocol;
	unsigned char action;
} entry;

//static __declspec(scratch) unsigned int pkt_count[8] = (__declspec(scratch) unsigned int *)SCR_PKT_COUNTER_ADDR;
//static __declspec(scratch) unsigned int byte_count[8] = (__declspec(scratch) unsigned int *)SCR_BYTE_COUNTER_ADDR;

void main()
{

////////////////////////////////////////////////////////////////////////////////
//
//	Signal and register definitions
//
////////////////////////////////////////////////////////////////////////////////

	SIGNAL 			sig_scr_get;				// signal for scratch get
	SIGNAL volatile	sig_next;					// signal for next ctx thread ordering
	SIGNAL			sig_scr_put;				// signal for scratch put
	SIGNAL_PAIR	sig_get_dram_header_pair;
	SIGNAL	volatile system_init_done;
	SIGNAL			counter_sig;
	SIGNAL			sram_sig_base1;
	
	__declspec(gp_reg) 	int temp;				// general purpose use
	__declspec(gp_reg) 	int port;				// input port number
	__declspec(sram_write_reg) unsigned	int txreq;		// tx request message to put on scratch rings
	__declspec(gp_reg) 	int port_out;			// output port to forward packet to
	__declspec(sram_read_reg) ring_data_t3 rdata;
	__declspec(dram_read_reg) ip_header  header;
	__declspec(shared gp_reg) unsigned int counter_p0, counter_p1, counter_p2, counter_p3;
	__declspec(gp_reg) unsigned int loop, default_rule;
	__declspec(gp_reg) unsigned int sip, dip, sp, dp, po;
	__declspec(gp_reg) unsigned int flow_id, ii;

	__declspec(sram_read_reg) entry 	sram_entry;

	__declspec(sram_write_reg) unsigned int pkt_size;

	//__declspec(shared scratch) unsigned int byte_count[8];
	//__declspec(shared scratch) unsigned int pkt_count[8];

	__declspec(sram_write_reg) unsigned int clear;

// This is a global signal that indicates that 
// system initialization is done

	__assign_relative_register((void *)&system_init_done, 10);

////////////////////////////////////////////////////////////////////////////////
//
//	Context common initializations
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//
//	Context specific initializations
//
////////////////////////////////////////////////////////////////////////////////


	if(ctx() == 0)
	{
		__wait_for_all(&system_init_done);
		counter_p0 = 0;
		counter_p1 = 0;
		counter_p2 = 0;
		counter_p3 = 0;

		//byte_count[0] = byte_count[1] = byte_count[2] = byte_count[3] = 0;
		//byte_count[4] = byte_count[5] = byte_count[6] = byte_count[7] = 0;

		//pkt_count[0] = pkt_count[1] = pkt_count[2] = pkt_count[3] = 0;
		//pkt_count[4] = pkt_count[5] = pkt_count[6] = pkt_count[7] = 0;

		clear = 0;

		for( ii=0; ii<16; ii++ )
			scratch_write( &clear,  (__declspec(scratch) unsigned int *)(SCR_BASE+(ii<<2)), 1, ctx_swap, &counter_sig);

		signal_same_ME(__signal_number(&sig_next), 1);
		__implicit_write((void *)&sig_next);
		signal_same_ME(__signal_number(&sig_next), 2);
		__implicit_write((void *)&sig_next);
		signal_same_ME(__signal_number(&sig_next), 3);
		__implicit_write((void *)&sig_next);
		signal_same_ME(__signal_number(&sig_next), 4);
		__implicit_write((void *)&sig_next);
		signal_same_ME(__signal_number(&sig_next), 5);
		__implicit_write((void *)&sig_next);
		signal_same_ME(__signal_number(&sig_next), 6);
		__implicit_write((void *)&sig_next);
		signal_same_ME(__signal_number(&sig_next), 7);
		__implicit_write((void *)&sig_next);
	}
	else
	{__wait_for_all(&sig_next);}

	if(ctx() == 0)
	{
		signal_same_ME(__signal_number(&sig_next), 0);
		__implicit_write((void *)&sig_next);	
	}

	signal_same_ME(__signal_number(&sig_scr_put), __ctx());
	__implicit_write(&sig_scr_put);	

////////////////////////////////////////////////////////////////////////////////
//
//	Main loop
//
////////////////////////////////////////////////////////////////////////////////
	
	while(1)
	{

			/// START //////////////////////////////////
			
			/// Wait1 //////////////////////////////////

			__wait_for_all(&sig_next, &sig_scr_put);

			scratch_get_ring(&rdata, (__declspec(scratch) unsigned int *)(POS_RX_RING_OUT<<2), 5, sig_done, &sig_scr_get);

			/// Signal1 ////////////////////////////////

			signal_same_ME(__signal_number(&sig_next), (__ctx())+1 );//sig_next_val);
			__implicit_write((void *)&sig_next);

			/// Wait2 //////////////////////////////////

			__wait_for_all(&sig_scr_get);

			if(rdata.dlBufHandle.value != 0 )
			{	
				dram_read(&header, (__declspec(dram) unsigned int *) (Dl_BufGetData(rdata.dlBufHandle) + rdata.offset ), 5, sig_done, &sig_get_dram_header_pair);
				__wait_for_all(&sig_get_dram_header_pair);

				sip = ( header.source_addr_high << 16 ) | header.source_addr_low;
				dip = ( header.destination_addr_high << 16 ) | header.destination_addr_low;
				sp = header.source_port;
				dp = header.destination_port;
				po = header.protocol;
	
				// Get Pkt Size
				pkt_size = rdata.buffer_size;

				// Decide Output Port
				port_out = 0;				
				 				
				// Get Flow ID
				if(   ( (sip>>31)==1 ) && ( (dip>>31)==0 )  ) flow_id = 0;
				else if( (dip>>31)==0 ) flow_id = 1;
				else if( (sip>>31)==1 ) flow_id = 2;
				else flow_id = 3;

				// Update Counter;
				scratch_incr( (__declspec(scratch) unsigned int *) (SCR_PKT_COUNTER_ADDR + (flow_id << 2) ) );
				scratch_add( &pkt_size, (__declspec(scratch) unsigned int *) (SCR_BYTE_COUNTER_ADDR + (flow_id << 2) ), ctx_swap, &counter_sig );


				if( port_out != header.ttl )
				{
					__asm __attribute(LITERAL_ASM)
					{ nop }
				}

				//port = rdata.input_port & 0xF;
				//port_out = port;
				
				txreq = (1<<31) | (port_out << 24) | ((unsigned int)(rdata.dlBufHandle.lw_offset) & 0xFFFFFF); 		

				if(port_out == 0) 
				{
					while(inp_state_test(inp_state_scr_ring6_status))
						ctx_swap();
					scratch_put_ring(&txreq, (__declspec(scratch) unsigned int *)(PACKET_TX_SCR_RING_0<<2), 1, sig_done, &sig_scr_put);
					continue;
				}
				if(port_out == 1) 
				{	while(inp_state_test(inp_state_scr_ring7_status))
						ctx_swap();
					scratch_put_ring(&txreq, (__declspec(scratch) unsigned int *)(PACKET_TX_SCR_RING_1<<2), 1, sig_done, &sig_scr_put);
					counter_p1++;
					continue;
				}
				if(port_out == 2) 
				{
					while(inp_state_test(inp_state_scr_ring8_status))
						ctx_swap();
					scratch_put_ring(&txreq, (__declspec(scratch) unsigned int *)(PACKET_TX_SCR_RING_2<<2), 1, sig_done, &sig_scr_put);
					counter_p2++;
					continue;
				}
				if(port_out == 3) 
				{
					while(inp_state_test(inp_state_scr_ring9_status))
						ctx_swap();
					scratch_put_ring(&txreq, (__declspec(scratch) unsigned int *)(PACKET_TX_SCR_RING_3<<2), 1, sig_done, &sig_scr_put);
					counter_p3++;
					continue;
				}
			}
			else
			{
				signal_same_ME(__signal_number(&sig_scr_put), __ctx());
				__implicit_write(&sig_scr_put);
			}
	}
}
